/***************************************************************************
                          Linesrv communication class
                             -------------------
    begin                : Mon Feb 12 2001
    copyright            : (C) 2001 by Peter Simonsson
                           (C) 2005 by Stefan Gehn
    email                : metz \AT\ gehn \DOT\ net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// Standard headers
#include <stdlib.h>

// Qt headers
#include <qstring.h>
#include <qtimer.h>
#include <qtcpsocket.h>

// App headers
#include "lcpio.h"
#include "lcpline.h"
#include "sendqpost.h"

//#define LCPIO_PACKET_DEBUG

LcpIO::LcpIO(QObject* parent) : QObject(parent)
{
	//qDebug("LcpIO::LcpIO()");
	// Socket
	mSocket = 0;
	connect(this, SIGNAL(cmdTimedOut()), this, SLOT(connectionLost()));

	// init sendQ
	m_sendQ = new QLinkedList<SendQPost *>();
	//m_sendQ->setAutoDelete(true);

	// init line list
	mLines = new QList<LcpLine *>();
	//mLines->setAutoDelete(true);

	LcpLine *defaultLine = new LcpLine();
	defaultLine->setLineId(0);

	mLines->append(defaultLine);
	m_selectedLine = mLines->first();

	// Create buffer for incoming packets
	m_inBuffer = (char*) malloc(get_max_lcp3_pack_size() * 3);
	m_received = 0;

	// AutoOnline
	m_autoOnline = false;

	// Line handling
	m_savedLineId = 0;
	m_allLinesReceived = false;

	// ISND channels handling
	m_waitChanUpdate = false;

	// Auto reconnect
	m_autoReconnectEnabled = false;
	m_autoReconnectDelay = 0;
	m_autoReconnectCount = 0;
	m_autoReconnectTries = 0;
	m_autoReconnectTimer = new QTimer(this);
	connect(m_autoReconnectTimer, SIGNAL(timeout()), this, SLOT(reconnect()));

	// Auto redial
	m_autoRedialEnabled = false;
	m_autoRedialDelay = 0;
	m_autoRedialCount = 0;
	m_autoRedialTries = 0;
	m_autoRedialTimer = new QTimer(this);
	connect(m_autoRedialTimer, SIGNAL(timeout()), this, SLOT(redial()));
	m_allowOffline = true;

	// Send timers
	m_resendTimer = new QTimer(this);
	m_keepAliveTimer = new QTimer(this);
	connect(m_resendTimer, SIGNAL(timeout()), this, SLOT(processSendQ()));
	connect(m_keepAliveTimer, SIGNAL(timeout()), this, SLOT(keepAlive()));
	// Client Status
	m_clientStatus = CLT_DISCON;
	// Wait for all lines to be received before we do anything else
	//connect(this, SIGNAL(allLinesReceived(QList<LcpLine *> *)), this, SLOT(postConnectProcessing()));
}

LcpIO::~LcpIO()
{
	//qDebug("LcpIO::~LcpIO()");

	if(mSocket)
	{
		//Debug("LcpIO::~LcpIO() disconnecting...");

		// first go offline if we aren't offline yet
		if(m_clientStatus == CLT_ONLINE || m_clientStatus == CLT_DIALING)
		{
			m_allowOffline = true;
			addToSendQ(CMD3_OFFLINE);
		}
		// Let's hope going offline worked out, we have to disconnect now in any case
		addToSendQ(CMD3_UNREG);

		// wait for a proper cleanup, or at least try to
		if (!mSocket->waitForDisconnected(5000))
		{
			qDebug("Connection to server did not shutdown properly, forcing disconnect");
			mSocket->abort(); // causes a disconnected() signal which also triggers cleanup()
		}
	}

	while (!m_sendQ->empty())
		delete m_sendQ->takeFirst();
	while (!mLines->empty())
		delete mLines->takeFirst();

	delete mLines;
	delete m_sendQ;
	delete m_inBuffer;
}

QList<LcpLine *>* LcpIO::lineList() const
{
	return mLines;
}

LcpLine* LcpIO::selectedLine() const
{
	return m_selectedLine;
}

void LcpIO::setAutoReconnect(bool use, int delay, int count)
{
	m_autoReconnectEnabled = use;
	m_autoReconnectDelay = delay;
	m_autoReconnectCount = count;
}

void LcpIO::setAutoRedial(bool use, int delay, int count)
{
	m_autoRedialEnabled = use;
	m_autoRedialDelay = delay;
	m_autoRedialCount = count;
}

void LcpIO::processCmd(void* incoming)
{
	struct t_lcp3_cmd* packet = static_cast<struct t_lcp3_cmd*>(incoming);
	void* infoPtr = static_cast<char*>(incoming) + sizeof(struct t_lcp3_cmd);

/*#ifdef LCPIO_PACKET_DEBUG
	qDebug("Received command %d", packet->cmd);
#endif*/

	// Send ack?
	if((packet->cmd < MIN_CBR3) && (packet->cmd != CMD3_ACK) && (packet->cmd != CMD3_UNREG))
		sendAck(packet->cmd);

	if(packet->info_is_valid)
	{
		//qDebug("got valid packet, current client status is %d", m_clientStatus);

		// connecting to server or going offline?
		if((m_clientStatus == CLT_NOAUTH) && (packet->client_stat == CLT_OFFLINE))
		{
			//qDebug("Requesting lines from server");
			addToSendQ(CMD3_LINESREQ); // Request lines from server
			m_autoReconnectTries = 0;
		}
		else if ((m_clientStatus > CLT_OFFLINE) && (packet->client_stat <= CLT_OFFLINE))
		{
			if(m_autoRedialEnabled && !m_allowOffline && (m_autoRedialTries < m_autoRedialCount))
			{
				m_autoRedialTimer->setSingleShot(true);
				m_autoRedialTimer->start(m_autoRedialDelay * 1000);
			}
			emit clientOffline();
		}

		// Clientstatus update?
		if (m_clientStatus != packet->client_stat)
		{
			m_clientStatus = packet->client_stat;
#ifdef LCPIO_PACKET_DEBUG
			qDebug("Client status changed, new status: %d", m_clientStatus);
#endif
			emit clientstatusChanged(m_clientStatus);

			if(m_clientStatus == CLT_ONLINE)
				m_autoRedialTries = 0;
		}

		if (m_allLinesReceived)
		{
			LcpLine* line = findLine(packet->line_id);

			// Linestatus update
			if(line && (line->status() != packet->line_stat))
			{
#ifdef LCPIO_PACKET_DEBUG
				qDebug("Line status for line %d is %d", packet->line_id, packet->line_stat);
#endif
				line->setStatus(packet->line_stat);
				emit lineStatusChanged(packet->line_stat, packet->line_id);

				// is it the selected line that changed status?
				if(m_selectedLine == line)
				{
#ifdef LCPIO_PACKET_DEBUG
					qDebug("Selected line changed its status");
#endif
					emit selectedLineStatusChanged(packet->line_stat);
				}

				emit numLinesOnlineChanged(LcpLine::numLinesOnline());
			}
		}
	} // END if(packet->info_is_valid)



	// process packet
	switch(packet->cmd)
	{
	case CMD3_AUTHBAD:
		{
			qDebug("Received 'bad auth' packet, login failed!");
			// TCP connection gets dropped by server so disconnect as well
			mSocket->close();
			//cleanup();
			emit badAuth();
			break;
		}

	case CMD3_CLIENTSTAT:
		// Nothing is needed here!
		break;

	case CMD3_LINE:
		{
			LcpLine* line = new LcpLine(static_cast<t_lcp3_info_line*>(infoPtr)->linename,
				/*mSocket->host()*/ mSocket->peerName());

			line->setStatus(CST_DOWN);
			line->setLineId(static_cast<t_lcp3_info_line*>(infoPtr)->line_id);
			line->setMaxThroughputUp(static_cast<t_lcp3_info_line*>(infoPtr)->capability_baud_up);
			line->setMaxThroughputDown(static_cast<t_lcp3_info_line*>(infoPtr)->capability_baud_down);
			line->setNumChannels(1);

			mLines->append(line);
#ifdef LCPIO_PACKET_DEBUG
			qDebug("Added line with id: %d", static_cast<t_lcp3_info_line*>(infoPtr)->line_id);
#endif
			if((mLines->count() - 1) == (static_cast<t_lcp3_info_line*>(infoPtr)->total_lines))
			{
#ifdef LCPIO_PACKET_DEBUG
				qDebug("All lines received");
#endif
				postConnectProcessing();
				//emit allLinesReceived(mLines);
			}
			break;
		}

	case CMD3_STATUS:
#ifdef LCPIO_PACKET_DEBUG
		qDebug("Received server status");
#endif
		emit serverStatus(*static_cast<t_lcp3_info_status*>(infoPtr));
		break;

	case CMD3_INVALID:
#ifdef LCPIO_PACKET_DEBUG
		qDebug("Server says client sent an invalid command");
#endif
		emit invalidCommand(static_cast<t_lcp3_info_ack*>(infoPtr)->ackcmd);
		break;

	case CMD3_ENCINFO:
		{
#ifdef LCPIO_PACKET_DEBUG
			qDebug("Received encryption info for login");
#endif
			struct t_lcp3_info_auth* infoAuth = new struct t_lcp3_info_auth;
			infoAuth->enctype = static_cast<t_lcp3_info_enckey*>(infoPtr)->enctype;

			switch(infoAuth->enctype)
			{
			case LCP3_ENCTYPE_OPEN: // No authentication!
				infoAuth->name[0] = 0;
				infoAuth->pass[0] = 0;
				break;

			case LCP3_ENCTYPE_SYM:
			case LCP3_ENCTYPE_ASYM:
				// These are unsupported at the moment
				qDebug("Unsupported authtype %d", infoAuth->enctype);
				// fallthrough

			case LCP3_ENCTYPE_NONE: // Plain text! (Default)
			default:
				qstrncpy(infoAuth->name, m_user.toLocal8Bit(), LCP3_MAX_NAME_LEN);
				qstrncpy(infoAuth->pass, m_passwd.toLocal8Bit(), LCP3_MAX_PWD_LEN);
				break;
			}

			addToSendQ(CMD3_AUTHREQ, infoAuth, sizeof(struct t_lcp3_info_auth));
			delete infoAuth;
			infoAuth = 0;
			break;
		}

	case CMD3_NOACCESS:
		emit accesserror(static_cast<t_lcp3_info_ack*>(infoPtr)->ackcmd);
		break;

	case CMD3_CHANFAIL:
		{
			m_waitChanUpdate = false;
			emit chanFail();
			break;
		}

	case CMD3_ACK:
		{
			int ackedcmd = static_cast<t_lcp3_info_ack*>(infoPtr)->ackcmd;
#ifdef LCPIO_PACKET_DEBUG
			qDebug("Received ACK for command: %d", ackedcmd);
#endif
			if(ackedcmd != CMD3_REFRESH)
				processAck(ackedcmd);
			break;
		}

	case CMD3_UNREG: // Disconnect from server
#ifdef LCPIO_PACKET_DEBUG
		qDebug("Received disconnect from server, closing down socket...");
#endif
		mSocket->disconnectFromHost(); //close();
		//cleanup();
		break;

	case CBR3_TPUT:
		{
			double clientCost = 0, lineCost = 0;

			LcpLine* line = findLine(packet->line_id);
			if(!line)
				break;

			t_lcp3_info_tput tput = *static_cast<t_lcp3_info_tput *>(infoPtr);

			line->calculateCost(tput, clientCost, lineCost);

			if(m_selectedLine == line)
			{
				if(line->autoOffline(tput))
				{
					emit selectedLineAutoOffline(tput);
					//offline();
				}
				emit selectedLineInfoUpdate(tput, clientCost, lineCost);
			}
			emit lineInfoUpdate(tput, clientCost, lineCost, packet->line_id);
			break;
		}

	case CBR3_LINESTAT:
		{
			// Num of ISDN channels used
			LcpLine* line = findLine(packet->line_id);
			if(!line)
				break;

			if(line->numChannels() != packet->line_channels)
			{
				line->setNumChannels(packet->line_channels);
				emit lineChanChanged(packet->line_channels);

				if(m_selectedLine == line)
				{
					emit selectedLineChanChanged(packet->line_channels);
					m_waitChanUpdate = false;
				}
			}
			else if(m_waitChanUpdate && (m_selectedLine == line))
			{
				emit chanFail();
				m_waitChanUpdate = false;
			}
			break;
		}

	case CBR3_SRVDEATH:
		{
			qDebug("Server is gone!, cleaning up");
			// The server has left the building... Time to clean up after the party!
			cleanup();

			if((m_autoReconnectTries <= m_autoReconnectCount) && m_autoReconnectEnabled)
			{
				m_autoReconnectTimer->setSingleShot(true);
				m_autoReconnectTimer->start(m_autoReconnectDelay * 1000);
			}
			else
			{
				emit connectionDown();
			}
			break;
		}

	case CBR3_ISDNACT:
		emit isdnAct(*static_cast<t_lcp3_info_isdn*>(infoPtr));
		break;

	default: // Unsupported commands
		qDebug("Received unsupported command: %d", packet->cmd);
	} // END switch()
}


void LcpIO::processAck(int cmd)
{
	SendQPost *cmdPacket = 0;
	bool isFound = false;
#ifdef LCPIO_PACKET_DEBUG
	qDebug("Processing ack for command %d, searching in sendQueue...", cmd);
#endif
	QLinkedList<SendQPost *>::iterator it;
	for (it = m_sendQ->begin(); it != m_sendQ->end(); ++it)
	{
		cmdPacket = (*it);
		if(cmdPacket && cmdPacket->getCmd() == cmd)
		{
			delete cmdPacket;
			m_sendQ->erase(it);
			isFound = true;
			break;
		}
	}

	if(!isFound)
		qDebug("Didn't find to-be-ACKed command %d in queue", cmd);
}


void LcpIO::addToSendQ(int cmd, void* info, int infoSize)
{
	SendQPost* temp = 0;

	temp = new SendQPost(cmd, info, infoSize);
	m_sendQ->append(temp);

#ifdef LCPIO_PACKET_DEBUG
	qDebug("Sending command %d", cmd);
#endif
	if(!temp->send(mSocket, m_selectedLine->lineId()))
	{
		emit cmdTimedOut();
	}
	else
	{
		// Activate resendTimer if it's inactive
		if(!m_resendTimer->isActive())
			m_resendTimer->start(RESENDINTERVAL);

		// Deactivate keepAliveTimer if it's active
		if(m_keepAliveTimer->isActive())
			m_keepAliveTimer->stop();
	}
}


void LcpIO::sendAck(int cmd)
{
	t_lcp3_info_ack infoAck;
#ifdef LCPIO_PACKET_DEBUG
	qDebug("Sending ACK for command: %d", cmd);
#endif
	infoAck.ackcmd = cmd;
	// Create a temporary SendQPost so we can use send!!
	SendQPost temp(CMD3_ACK, &infoAck, sizeof(t_lcp3_info_ack));

	if(!temp.send(mSocket, m_selectedLine->lineId()))
		emit cmdTimedOut();
}


void LcpIO::cleanup()
{
	//qDebug("Cleaning up, BEGIN");
	m_clientStatus = CLT_DISCON;
	emit clientstatusChanged(m_clientStatus);

	QList<LcpLine *>::iterator it;
	QList<LcpLine *>::iterator end = mLines->end();
	for (it = mLines->begin(); it != end; ++it)
		emit lineStatusChanged(CST_DOWN, (*it)->lineId());

	emit selectedLineStatusChanged(CST_DOWN);
	emit numLinesOnlineChanged(0);

	m_resendTimer->stop();
	m_keepAliveTimer->stop();

	// postpone socket deletion, we might be called from a mSocket signal
	mSocket->deleteLater();
	mSocket = 0;

	while (!m_sendQ->empty())
		delete m_sendQ->takeFirst();
	while (!mLines->empty())
		delete mLines->takeFirst();

	m_received = 0;

	LcpLine* tmp = new LcpLine();
	tmp->setLineId(0);
	mLines->append(tmp);
	m_selectedLine = mLines->first();

	m_allLinesReceived = false;
	//qDebug("Cleaning up, END");
}


LcpLine* LcpIO::findLine(int lineId)
{
	if(m_selectedLine->lineId() == lineId)
	{
		return m_selectedLine;
	}

	QList<LcpLine *>::iterator it;
	QList<LcpLine *>::iterator end = mLines->end();
	for (it = mLines->begin(); it != end; ++it)
	{
		LcpLine *line = *it;
		if(line && line->lineId() == lineId)
			return line;
	}

	return 0;
}


void LcpIO::receiveData()
{
	int size = 0;
	int packetsize = 0;
	int minsize = sizeof(t_lcp3_cmd);
	int available = get_max_lcp3_pack_size() * 3 - m_received; // Size of unused area
	char* ptr = (char *) m_inBuffer + m_received; // Pointer to unused area
	char* temp = 0;

#ifdef LCPIO_PACKET_DEBUG
	qDebug("LcpIO::receiveData() called, reading up to %d bytes from socket...", available);
#endif

	size = mSocket->read(ptr, available);

#ifdef LCPIO_PACKET_DEBUG
	qDebug("Read %d bytes from socket", size);
#endif

	// Check if OK
	if(size > 0)
	{
		m_received += size;
#ifdef LCPIO_PACKET_DEBUG
		qDebug("New buffersize: %d", m_received);
#endif
		// Process all received packets
		while(m_received >= minsize)
		{
			packetsize = get_lcp3_pack_size(((t_lcp3_cmd*) m_inBuffer)->cmd);
#ifdef LCPIO_PACKET_DEBUG
			qDebug("Received command %d, size is %d", ((t_lcp3_cmd*) m_inBuffer)->cmd, packetsize);
#endif
			if(m_received >= packetsize)
			{
				// Process the received packet
				temp = (char*)malloc(packetsize);
				memcpy(temp, m_inBuffer, packetsize);
				// Remove processed size
				m_received -= packetsize;
#ifdef LCPIO_PACKET_DEBUG
				qDebug("Bytes left in buffer: %d", m_received);
#endif
				// Move remaining data to beginning of inBuffer
				if(m_received > 0)
				{
					ptr = (char*)m_inBuffer + packetsize;
					memmove(m_inBuffer, ptr, m_received);
				}
				else if(m_received < 0)
				{
					qDebug("Ooops! Memory management went bad! Received = %d", m_received);
					m_received = 0;
				}

				// Actually handle incoming
				processCmd(temp);

				free(temp);
				temp = 0;
			}
			else
			{
#ifdef LCPIO_PACKET_DEBUG
				qDebug("Received incomplete packet, waiting for more");
#endif
				// The whole packet hasn't been received yet!
				break;
			}
		} // END while()

		// socket still got data that did not fit into our own buffer, read more
		if (mSocket && mSocket->bytesAvailable() > 0)
		{
			qDebug("Still got data to read from socket, calling myself...");
			QTimer::singleShot(0, this, SLOT(receiveData()));
		}

	}
	else // size <= 0
	{
		qDebug("Error! Couldn't read from socket.");
	}
}


void LcpIO::keepAlive()
{
	// we are (still) not logged in but already sending
	// keepalive packets, that can't be right
	if (m_clientStatus < CLT_OFFLINE)
	{
		qDebug("Connection foobar'd, wanted to send keepalive while authenticating");
		addToSendQ(CMD3_UNREG);
	}
	else
	{
#ifdef LCPIO_PACKET_DEBUG
		qDebug("Sending keepalive command (%d)", CMD3_REFRESH);
#endif
		SendQPost temp(CMD3_REFRESH);
		if(!temp.send(mSocket, m_selectedLine->lineId()))
			emit cmdTimedOut();
	}
}


void LcpIO::processSendQ()
{
	SendQPost *temp;

	QLinkedList<SendQPost *>::iterator it;
	for (it = m_sendQ->begin(); it != m_sendQ->end(); ++it)
	{
		temp = (*it);

		if(temp->isTimedOut() && !temp->noTryLeft())
		{
#ifdef LCPIO_PACKET_DEBUG
			qDebug("Re-sending packet with command %d", temp->getCmd());
#endif
			if(!temp->send(mSocket, m_selectedLine->lineId()))
			{
				emit cmdTimedOut();
				return;
			}

		}
		else if(temp->noTryLeft())
		{
			delete temp;
			qDebug("Packet containing command %d timed out and no retry left!", temp->getCmd());
			m_sendQ->erase(it); // Remove current item in list
			emit cmdTimedOut();
			return;
		}
	}

	// If the Q is empty we don't need to process it,
	// so we stop the resendTimer.
	if(m_sendQ->isEmpty())
	{
		m_resendTimer->stop();
		m_keepAliveTimer->start(REFRESHINTERVAL);
	}
}


/*void LcpIO::slotSocketStateChanged(QAbstractSocket::SocketState state)
{
	switch(state)
	{
	case QAbstractSocket::UnconnectedState:
		qDebug("UNCONNECTED");
		break;
	case QAbstractSocket::HostLookupState:
		qDebug("HOSTLOOKUP");
		break;
	case QAbstractSocket::ConnectingState:
		qDebug("CONNECTING");
		break;
	case QAbstractSocket::ConnectedState:
		qDebug("CONNECTED");
		break;
	case QAbstractSocket::BoundState:
		qDebug("BOUND");
		break;
	case QAbstractSocket::ClosingState:
		qDebug("CLOSING");
		break;
	case QAbstractSocket::ListeningState:
		qDebug("LISTENING");
		break;
	}
}*/


void LcpIO::slotSocketConnected()
{
	qDebug("LcpIO::slotConnected()");
	addToSendQ(CMD3_PUBKEYREQ);
}


void LcpIO::slotSocketError(QAbstractSocket::SocketError /*err*/)
{
	qDebug("LcpIO::slotSocketError(), error description: %s",
		mSocket->errorString().toLocal8Bit().constData());

	connectionLost();
}


void LcpIO::connectionLost()
{
	qDebug("LcpIO::connectionLost()");

	// Let's pretend that we are disconnected now
	mSocket->abort();
	cleanup();

	if(m_autoReconnectCount && (m_autoReconnectTries <= m_autoReconnectCount))
	{
		m_autoReconnectTimer->setSingleShot(true);
		m_autoReconnectTimer->start(m_autoReconnectDelay * 1000);
	}
	else
	{
		emit connectionDown();
	}
}


void LcpIO::postConnectProcessing()
{
	//qDebug("LcpIO::postConnectProcessing()");

	m_allLinesReceived = true;
	emit allLinesReceived(mLines);

	if(m_savedLineId > 0)
	{
		qDebug("Restoring saved line-id");
		setSelectedLine(m_savedLineId);
	}
	else
	{
		qDebug("Selecting first available line");
		// Line at index 0 is default, line at index 1
		// is the first "real" one we received from the server
		if (mLines->count() > 1)
		{
			setSelectedLine(mLines->at(1)->lineId());
			/*
			m_selectedLine = mLines->at(1);
			addToSendQ(CMD3_LINESEL);
			*/
		}
	}

	addToSendQ(CMD3_GETALL); // Get status for all lines

	// Auto online?
	if(m_autoOnline)
	{
		online();
		//addToSendQ(CMD3_ONLINE);
	}

	/*qDebug("Selected Line: %s [%d]", m_selectedLine->name().toLatin1().data(),
		m_selectedLine->lineId());

	emit selectedLineChanged(m_selectedLine);
	*/
}


void LcpIO::reconnect()
{
	qDebug("LcpIO::reconnect()");

	m_autoReconnectTries++;
	connectInternal();
}


void LcpIO::redial()
{
	qDebug("LcpIO::redial()");

	m_autoRedialTries++;
	addToSendQ(CMD3_ONLINE);

	// Use two ISDN channels as default !!
	if(m_selectedLine->useTwoChanDefault())
		addToSendQ(CMD3_CHANADD);
}


void LcpIO::connectToServer(const QString host, const uint port,
							const QString user, const QString passwd)
{
	//qDebug("LcpIO::connectToServer()");
	if(m_clientStatus <= CLT_NOAUTH)
	{
		m_user = user;
		m_passwd = passwd;
		m_host = host;
		m_port = port;
		m_autoReconnectTries = 0;

		connectInternal();
	}
	else
		qDebug("Trying to connect while already in connected state");
}


void LcpIO::connectInternal()
{
	delete mSocket;

	/*qDebug("LcpIO::connectInternal(); Connecting to '%s:%d' ...",
		m_host.toLocal8Bit().constData(), m_port);
	*/

	mSocket = new QTcpSocket(this);
	connect(mSocket, SIGNAL(readyRead()), this, SLOT(receiveData()));
	connect(mSocket, SIGNAL(disconnected()), this, SLOT(cleanup()));
	connect(mSocket, SIGNAL(connected()), this, SLOT(slotSocketConnected()));
	connect(mSocket, SIGNAL(error(QAbstractSocket::SocketError)),
		this, SLOT(slotSocketError(QAbstractSocket::SocketError)));
	/*connect(mSocket, SIGNAL(stateChanged(QAbstractSocket::SocketState)),
		this, SLOT(slotSocketStateChanged(QAbstractSocket::SocketState)));*/

	mSocket->connectToHost(m_host, m_port);
}


void LcpIO::disconnectFromServer()
{
	if(m_clientStatus == CLT_OFFLINE)
		addToSendQ(CMD3_UNREG);
	else
		qDebug("Tried to disconnect from server while client is not offline, ignoring");
}


void LcpIO::online()
{
	// If we are offline then go online
	if(m_clientStatus == CLT_OFFLINE)
	{
		m_allowOffline = false;
		m_autoRedialTries = 0;
		addToSendQ(CMD3_ONLINE);

		// Use two ISDN channels as default !!
		if(m_selectedLine->useTwoChanDefault())
			addToSendQ(CMD3_CHANADD);
	}
	else
		qDebug("Tried to go online while client is not offline, ignoring");
}


void LcpIO::offline()
{
	// If we are online then go offline
	if(m_clientStatus == CLT_ONLINE || m_clientStatus == CLT_DIALING)
	{
		m_allowOffline = true;
		addToSendQ(CMD3_OFFLINE);

		// Are we aborting a online request...
		if(m_clientStatus == CLT_DIALING)
			emit clientstatusChanged(CLT_ABORTING);
	}
	else
		qDebug("Tried to go offline while client is not online/dialing, ignoring");
}


void LcpIO::setAutoOnline(bool ao)
{
	m_autoOnline = ao;
}


void LcpIO::setSelectedLine(int lineId)
{
	//qDebug("LcpIO::setSelectedLine(%d)", lineId);
	if (m_allLinesReceived)
	{
		LcpLine *line = findLine(lineId);
		if (line)
		{
			m_selectedLine = line;
			addToSendQ(CMD3_LINESEL);

			/*qDebug("Selected Line is now  %s [%d]",
				m_selectedLine->name().toLatin1().data(),
				m_selectedLine->lineId());*/

			emit selectedLineChanged(m_selectedLine);
		}
		else
			qDebug("Could not find line-id in list of lines!");
	}
	else
	{
		m_savedLineId = lineId;
	}
}


void LcpIO::getServerStatus()
{
	addToSendQ(CMD3_STATREQ);
}


void LcpIO::toggleLineLock()
{
	if(m_selectedLine->status() == CST_LOCKED_UP_CLT)
		addToSendQ(CMD3_UNLOCK);
	else if(m_selectedLine->status() == CST_UP_SERVER)
		addToSendQ(CMD3_LOCK);
}


void LcpIO::toggleISDNChan()
{
	if(m_selectedLine->numChannels() == 1)
		addToSendQ(CMD3_CHANADD);
	else
		addToSendQ(CMD3_CHANRM);

	m_waitChanUpdate = true;
}


void LcpIO::shutdownServer()
{
	addToSendQ(CMD3_SHUTDOWN);
}
