/***************************************************************************
                          lcpio.h  -  description
                             -------------------
    begin                : Mon Feb 12 2001
    copyright            : (C) 2001 by Peter Simonsson
                           (C) 2005 by Stefan Gehn
    email                : metz \AT\ gehn \DOT\ net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef LCPIO_H
#define LCPIO_H

// Qt headers
#include <qobject.h>
#include <qlinkedlist.h>
#include <qlist.h>
#include <qabstractsocket.h>

// App headers
#include "lcp3.h"
#include "misc.h"

// Intervals
#define RESENDINTERVAL  3000  // Interval for resending commands
#define REFRESHINTERVAL 30000 // Interval for sending of REFRESH command

class LcpLine;
class SendQPost;
class QTcpSocket;
class QString;
class QTimer;

/**
 * Handles communication with linesrv thru LCP 3.3 protocol
 * @author Peter Simonsson
 */

class LcpIO : public QObject
{
Q_OBJECT

public:
	LcpIO(QObject* parent = 0);
	~LcpIO();

	//! @return A pointer to the line list
	QList<LcpLine *>* lineList() const;

	//! @return A pointer to the selected line
	LcpLine* selectedLine() const;

	//! setup auto reconnect
	//! @param use if auto reconnect should be used
	//! @param delay the delay between reconnects in seconds
	//! @param count number of reconnects
	void setAutoReconnect(bool use, int delay, int count);

	//! setup auto redial
	//! @param use if auto redial should be used
	//! @param delay the delay between redials in seconds
	//! @param count number of redials
	void setAutoRedial(bool use, int delay, int count);

public slots:
	//! Connect to linesrv
	//! @param host domainname or ip address of the server
	//! @param port port to connect to
	//! @param user username
	//! @param passwd password
	void connectToServer(const QString host, const uint port,
		const QString user, const QString passwd);

	//! disconnect from linesrv
	void disconnectFromServer();

	//! send online request to linesrv
	void online();

	//! send offline request to linesrv
	void offline();

	//! enable automatic online when client is connected
	//! @param ao state of automatic online
	void setAutoOnline(bool ao);

	//! select a line to use
	//! @param lineId line id of line to use
	void setSelectedLine(int lineId);

	//! get the servers status
	void getServerStatus();

	//! toggle locking of line on/off
	void toggleLineLock();

	//! toggle between one and two ISDN channels
	void toggleISDNChan();

	//! request the server to be shut down
	void shutdownServer();

signals:
	//! emited when all lines were received
	void allLinesReceived(QList<LcpLine *> *lines);

	//! emited when selected line changes
	void selectedLineChanged(LcpLine *);

	//! emited when a invalid command has been sent to the server
	void invalidCommand(int);

	//! emited when the client tries to use a restricted command
	void accesserror(int);

	//! emited if authentication fails
	void badAuth();

	//! emited when a info update comes for the selected line
	void selectedLineInfoUpdate(t_lcp3_info_tput, double clientCost, double lineCost);

	//! emited when auto offline kicks in for the selected line
	void selectedLineAutoOffline(t_lcp3_info_tput);

	//! emited whenever a info update is received
	void lineInfoUpdate(t_lcp3_info_tput, double, double, int);

	//! emited when there is an incomming call (ISDN only)
	void isdnAct(t_lcp3_info_isdn);

	//! emited whenever the clientstatus changes
	void clientstatusChanged(int);

	//! emited when connection to server is lost
	void connectionDown();

	//! emited when a server status update is received
	void serverStatus(t_lcp3_info_status);

	//! emited when status for any line changes
	void lineStatusChanged(int, int);

	//! emited when status for the selected line changes
	void selectedLineStatusChanged(int);

	//! emited when ISDN channel manipulation failed
	void chanFail();

	//! emited when the number of ISDN channels change for any line
	void lineChanChanged(int);

	//! emitted when the number of ISDN channels change for the selected line
	void selectedLineChanChanged(int);

	//! emited when the number of lines online changed
	void numLinesOnlineChanged(int);

	//! emited when a command in the sendQ times out
	void cmdTimedOut();

	//! emited when the client is going offline
	//! @note clientstatusChanged(int) is also emitted in this case.
	void clientOffline();

protected:
	//! Process received LCP packet
	//! @param incoming packet to process
	void processCmd(void* incoming);

	//! Remove acknowledged command from sendqueue
	//! @param cmd command to remove
	void processAck(int cmd);

	//! Add command to sendqueue
	//! @param cmd LCP command to send
	//! @param info info struct for command
	//! @param infoSize size of info
	void addToSendQ(int cmd, void* info = 0, int infoSize = 0);

	//! Send acknowledgement for received command
	//! @param cmd command to acknowledge
	void sendAck(int cmd);

	//! Find the line corresponding to line id
	//! @param lineId lineId for the line
	//! @return pointer to the line or 0 if the line was not found
	LcpLine* findLine(int lineId);

	void connectInternal();

protected slots:
	//! clean up after the connection is taken down
	void cleanup();

	//! receive data from server
	void receiveData();

	//! keep server connection alive
	void keepAlive();

	//! check if we need to (re)send command
	void processSendQ();

	void slotSocketError(QAbstractSocket::SocketError err);

	//void slotSocketStateChanged(QAbstractSocket::SocketState state);

	void slotSocketConnected();

	//! clean up after broken connection
	void connectionLost();

	//! called when all lines are received
	void postConnectProcessing();

	//! called when we need to reconnect
	void reconnect();

	//! called when we need to redial
	void redial();

private: // Only Variables
	QTcpSocket* mSocket;
	QTimer* m_resendTimer;
	QTimer* m_keepAliveTimer;
	char* m_inBuffer;
	int m_received;
	QLinkedList<SendQPost*>* m_sendQ;
	LcpLine* m_selectedLine;
	int m_clientStatus;
	QString m_host;
	uint m_port;
	QString m_user;
	QString m_passwd;
	bool m_autoOnline;
	int m_savedLineId;
	bool m_allLinesReceived;
	bool m_waitChanUpdate;
	bool m_autoReconnectEnabled;
	int m_autoReconnectDelay;
	int m_autoReconnectCount;
	int m_autoReconnectTries;
	QTimer* m_autoReconnectTimer;
	bool m_autoRedialEnabled;
	int m_autoRedialDelay;
	int m_autoRedialCount;
	int m_autoRedialTries;
	QTimer* m_autoRedialTimer;
	bool m_allowOffline;
	QList<LcpLine *> *mLines;
};

#endif
