/***************************************************************************
                          lcpline.cpp  -  description
                             -------------------
    begin                : Thu Oct 18 2001
    copyright            : (C) 2001 by Peter Simonsson
    email                : psn@linux.se
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lcpline.h"
#include "lcp3.h"

#include <qsettings.h>
#include <qapplication.h>

// Define static variables
bool LcpLine::s_autoOfflineDisabled = false;
int LcpLine::s_numLinesOnline = 0;

int LcpLine::numLinesOnline() { return s_numLinesOnline; }

bool LcpLine::autoOfflineDisabled() { return s_autoOfflineDisabled; }
void LcpLine::setAutoOfflineDisabled(bool disable) { s_autoOfflineDisabled = disable; }


// ----------------------------------------------------------------------------

LcpLine::LcpLine(const QString &name, const QString &host)
{
	m_host = host;
	m_name = name;
	m_numChannels = 1; // one channel is standard
	m_activityTime = 0;
	m_status = CST_DOWN;
	loadSettings();
}

LcpLine::~LcpLine()
{
	saveSettings();
	s_numLinesOnline = 0; // TODO: check if that's correct
}

QString LcpLine::host() const { return m_host; }

uint LcpLine::maxThroughputUp() const { return m_maxThroughputUp; }
void LcpLine::setMaxThroughputUp(uint m) { m_maxThroughputUp = m; }

uint LcpLine::maxThroughputDown() const { return m_maxThroughputDown; }
void LcpLine::setMaxThroughputDown(uint m) { m_maxThroughputDown = m; }

LCPIO::OfflineInfo LcpLine::offlineInfo() const { return m_offlineInfo; }
void LcpLine::setOfflineInfo(const LCPIO::OfflineInfo info) { m_offlineInfo = info; }

int LcpLine::numChannels() const { return m_numChannels; }
void LcpLine::setNumChannels(int num)
{
	m_maxThroughputUp = (m_maxThroughputUp/m_numChannels)*num; // set to new throughput
	m_maxThroughputDown = (m_maxThroughputDown/m_numChannels)*num; // set to new throughput
	m_numChannels = num;
}

LCPIO::CostInfo LcpLine::costInfo() const { return m_costInfo; }
void LcpLine::setCostInfo(const LCPIO::CostInfo info) { m_costInfo = info; }

bool LcpLine::useDefaultSettings() const { return m_useDefaultSettings; }
void LcpLine::setUseDefaultSettings(bool use) { m_useDefaultSettings = use; }

QString LcpLine::name() const { return (m_name.isNull() ? QApplication::translate("LCPIO", "Default", "Name of default line") : m_name); }
void LcpLine::setName(const QString name) { m_name = name; }

bool LcpLine::useTwoChanDefault() const { return m_useTwoChanDefault; }
void LcpLine::setUseTwoChanDefault(bool use) { m_useTwoChanDefault = use; }

bool LcpLine::onlineScriptActive() const { return m_onlineScriptActive; }
void LcpLine::setOnlineScriptActive(bool active) { m_onlineScriptActive = active; }

bool LcpLine::offlineScriptActive() const { return m_offlineScriptActive; }
void LcpLine::setOfflineScriptActive(bool active) { m_offlineScriptActive = active; }

const QString &LcpLine::onlineScript() const { return m_onlineScript; }
void LcpLine::setOnlineScript(const QString &path) { m_onlineScript = path; }

const QString &LcpLine::offlineScript() const { return m_offlineScript; }
void LcpLine::setOfflineScript(const QString &path) { m_offlineScript = path; }

int LcpLine::lineId() const { return m_lineId; }
void LcpLine::setLineId(int id) { m_lineId = id; }

int LcpLine::status() const { return m_status; }

void LcpLine::setStatus(int s)
{
	if((m_status == CST_UP_SERVER) ||
		(m_status == CST_UP_USER) ||
		(m_status == CST_LOCKED_UP_CLT) ||
		(m_status == CST_LOCKED_UP_TIMED))
	{
		if((s != CST_UP_SERVER) ||
			(s != CST_UP_USER) ||
			(s != CST_LOCKED_UP_CLT) ||
			(s != CST_LOCKED_UP_TIMED))
		{
			s_numLinesOnline--;
			qDebug("s_numLinesOnline now at %d", s_numLinesOnline);
		}
	}
	else if((m_status != CST_UP_SERVER) ||
		(m_status != CST_UP_USER) ||
		(m_status != CST_LOCKED_UP_CLT) ||
		(m_status != CST_LOCKED_UP_TIMED))
	{
		if((s == CST_UP_SERVER) ||
			(s == CST_UP_USER) ||
			(s == CST_LOCKED_UP_CLT) ||
			(s == CST_LOCKED_UP_TIMED))
		{
			s_numLinesOnline++;
			qDebug("s_numLinesOnline now at %d", s_numLinesOnline);
		}
	}

	m_status = s;
}


void LcpLine::calculateCost(struct t_lcp3_info_tput info, double& client, double& line)
{
	double fee = m_costInfo.connectionFee;
	double cost = m_costInfo.cost;

	if(m_costInfo.costCalcActive && m_costInfo.costType == LCPIO::Time)
	{
		if((m_status == CST_UP_SERVER) ||
			(m_status == CST_UP_USER) ||
			(m_status == CST_LOCKED_UP_CLT) ||
			(m_status == CST_LOCKED_UP_TIMED))
		{
			client = (fee * m_numChannels) + ((static_cast<int>(info.client / 60) + 1) * cost * m_numChannels);
		}
		else
		{
			client = 0;
		}

		line = (fee * m_numChannels) + ((static_cast<int>(info.line / 60) + 1) * cost * m_numChannels);
	}
}


bool LcpLine::autoOffline(struct t_lcp3_info_tput info)
{
	bool time, down, up, maxDown, maxUp;
	time = down = up = maxDown = maxUp = false;

	if(m_offlineInfo.autoOfflineActive && !s_autoOfflineDisabled)
	{
		if(m_offlineInfo.timeActive && (info.client >= m_offlineInfo.time))
			time = true;
		else if(!m_offlineInfo.timeActive && (m_offlineInfo.logic == LCPIO::AND))
			time = true;

		if(m_offlineInfo.downloadActive && (info.dns <= m_offlineInfo.download))
			down = true;
		else if(!m_offlineInfo.downloadActive && (m_offlineInfo.logic == LCPIO::AND))
			down = true;

		if(m_offlineInfo.uploadActive && (info.ups <= m_offlineInfo.upload))
			up = true;
		else if(!m_offlineInfo.uploadActive && (m_offlineInfo.logic == LCPIO::AND))
			up = true;

		if(m_offlineInfo.maxDownloadActive && (info.dnt >= m_offlineInfo.maxDownload))
			maxDown = true;
		else if(!m_offlineInfo.maxDownloadActive && (m_offlineInfo.logic == LCPIO::AND))
			maxDown = true;

		if(m_offlineInfo.maxUploadActive && (info.upt >= m_offlineInfo.maxUpload))
			maxUp = true;
		else if(!m_offlineInfo.maxUploadActive && (m_offlineInfo.logic == LCPIO::AND))
			maxUp = true;

		if(m_offlineInfo.mode == LCPIO::Activity)
		{
			if(m_offlineInfo.logic == LCPIO::AND)
			{
				if(down && up && maxDown && maxUp)
				{
					if(m_activityTime == 0)
						m_activityTime = info.client;
				}
				else
				{
					m_activityTime = 0;
				}
			}
			else if(m_offlineInfo.logic == LCPIO::OR)
			{
				if(down || up || maxDown || maxUp)
				{
					if(m_activityTime == 0)
						m_activityTime = info.client;
				}
				else
				{
					m_activityTime = 0;
				}
			}

			return (((info.client - m_activityTime) >= m_offlineInfo.time) && (m_activityTime != 0));
		}
		else if(m_offlineInfo.mode == LCPIO::Normal)
		{
			if(m_offlineInfo.logic == LCPIO::AND)
			{
				return (time && down && up && maxDown && maxUp);
			}
			else if(m_offlineInfo.logic == LCPIO::OR)
			{
				return (time || down || up || maxDown || maxUp);
			}
		}
	}

	return false;
}


void LcpLine::loadSettings()
{
	QSettings config(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());

	if(m_name.isNull()) // Default line
	{
		m_useDefaultSettings = true;
	}
	else
	{
		config.beginGroup(m_host + ":" + name());
		m_useDefaultSettings = config.value("Use Default", true).toBool();
		config.endGroup();
	}

	config.beginGroup(m_useDefaultSettings ? "__Default__" : (m_host + ":" + name()));

	m_offlineInfo.autoOfflineActive = config.value("Use Auto-offline", false).toBool();
	m_offlineInfo.timeActive = config.value("Use Time Limit", false).toBool();
	m_offlineInfo.time = config.value("Time Limit", 60).toUInt();
	m_offlineInfo.downloadActive = config.value("Use Download Limit", false).toBool();
	m_offlineInfo.download = config.value("Download Limit", 1000).toUInt();
	m_offlineInfo.uploadActive = config.value("Use Upload Limit", false).toBool();
	m_offlineInfo.upload = config.value("Upload Limit", 100).toUInt();
	m_offlineInfo.maxDownloadActive = config.value("Use Max Download Limit", false).toBool();
	m_offlineInfo.maxDownload = config.value("Max Download Limit", 20480).toUInt();
	m_offlineInfo.maxUploadActive = config.value("Use Max Upload Limit", false).toBool();
	m_offlineInfo.maxUpload = config.value("Max Upload Limit", 2048).toUInt();
	m_offlineInfo.logic = static_cast<LCPIO::LogicType> (config.value("Logic", LCPIO::OR).toInt());
	m_offlineInfo.mode = static_cast<LCPIO::ModeType> (config.value("Mode", LCPIO::Normal).toInt());
	m_costInfo.costCalcActive = config.value("Use Cost Calculation", false).toBool();
	m_costInfo.cost = config.value("Cost", 0.0).toDouble();
	m_costInfo.connectionFee = config.value("Connection Fee", 0.0).toDouble();
	m_costInfo.costType = static_cast<LCPIO::CostType> (config.value("Cost Type", LCPIO::Time).toInt());
	m_useTwoChanDefault = config.value("Default To Two Channels", false).toBool();
	m_onlineScriptActive = config.value("Use Online Script", false).toBool();
	m_onlineScript = config.value("Online Script", "").toString();
	m_offlineScriptActive = config.value("Use Offline Script", false).toBool();
	m_offlineScript = config.value("Offline Script", "").toString();

	config.endGroup();
}


void LcpLine::saveSettings()
{
	QSettings config(QSettings::IniFormat, QSettings::UserScope, qApp->organizationName(), qApp->applicationName());

	bool oldDefaultSetting = m_useDefaultSettings;
	if(m_name.isNull()) // Default line
	{
		config.beginGroup("__Default__");
	}
	else
	{
		config.beginGroup(m_host + ":" + name());
		config.setValue("Use Default", m_useDefaultSettings);

		if(m_useDefaultSettings)
		{
			// no need to save settings, we take them from default line
			config.endGroup();
			return;
		}
	}

	config.setValue("Use Auto-offline", m_offlineInfo.autoOfflineActive);
	config.setValue("Use Time Limit", m_offlineInfo.timeActive);
	config.setValue("Time Limit", m_offlineInfo.time);
	config.setValue("Use Download Limit", m_offlineInfo.downloadActive);
	config.setValue("Download Limit", m_offlineInfo.download);
	config.setValue("Use Upload Limit", m_offlineInfo.uploadActive);
	config.setValue("Upload Limit", m_offlineInfo.upload);
	config.setValue("Use Max Download Limit", m_offlineInfo.maxDownloadActive);
	config.setValue("Max Download Limit", m_offlineInfo.maxDownload);
	config.setValue("Use Max Upload Limit", m_offlineInfo.maxUploadActive);
	config.setValue("Max Upload Limit", m_offlineInfo.maxUpload);
	config.setValue("Logic", m_offlineInfo.logic);
	config.setValue("Mode", m_offlineInfo.mode);
	config.setValue("Use Cost Calculation", m_costInfo.costCalcActive);
	config.setValue("Cost", m_costInfo.cost);
	config.setValue("Connection Fee", m_costInfo.connectionFee);
	config.setValue("Cost Type", m_costInfo.costType);
	config.setValue("Default To Two Channels", m_useTwoChanDefault);
	config.setValue("Use Online Script", m_onlineScriptActive);
	config.setValue("Online Script", m_onlineScript);
	config.setValue("Use Offline Script", m_offlineScriptActive);
	config.setValue("Offline Script", m_offlineScript);

	config.endGroup();
}
